"""
sum with double precision
"""
function sum_double(x)
	return sum(Float64.(x))
end

"""
`xs, fs, rs, out = gm_restart(x0, f_cost, f_grad, f_L;
	f_mu=0, mom=0, restart=0, bsig=1, niter=10, seq=1, fun=?)`

GM / FGM / OGM, with (or without) restart, for smooth optimization

# In
* `x0` initial guess
* `f_cost` function for computing the cost function value f(x)  
	for efficiency, provide `f_cost` only if `restart == 1`
* `f_grad` function for computing the gradient of f(x)
* `f_L` Lipschitz constant of the gradient of f(x)

# Options
* `f_mu`	strong convexity parameter of f(x)
	if `f_mu > 0`, (alpha, beta_k, gamma_k) is chosen based on Table 1 in [KF17]
* `mom`	momentum option
	0: GM - beta_k = gamma_k = 0
	1: FGM - gamma_k = 0
	2: OGM
* `restart` restart option
		0: no restart
		1: function restart, see [OC15, KF17]
		2: gradient restart, see [OC15, KF17]
* `bsig`	gradient "gamma" decrease option (value within [0 1]),
		see σ̄ (&sigma;&#772;) in [KF18]
* `niter`	number of iterations
* `seq` output iterates
		0: last primary iterate `y_niter`
		1: primary iterates {y_k} (default, for historical reasons)
		2: secondary iterates {x_k}
* `fun`: fun(iter,xk) default just 0

# Out
* `xs`	iterates
* `fs`	cost function values
* `rs`	iterations where restarted
* `out` [niter+1] evaluations of fun(iter, xk)

Copyright 2017-3-31, Donghwan Kim and Jeff Fessler, University of Michigan
2018-08-12 Julia 0.7.0
2019-02-11 add user fun() option (for timing etc.)
"""
function gm_restart(x0, f_cost, f_grad, f_L;
	f_mu=0, mom=0, restart=0, bsig=1, niter=10, seq=1,
	fun::Function = (x,iter) -> 0)

if isnothing(f_cost) # provide dummy default
	if restart == 1
		error("restart requires f_cost")
	end
	f_cost = x -> 0
end

L = f_L
mu = f_mu
q = mu/L
rs = [] # restart iterations

# initialize parameters
told = 1
sig = 1

# initialize x
xold = x0
yold = x0
fcostold = f_cost(x0)
fgradold = zeros(size(x0)) # dummy

# save initial (if saving all)
if seq != 0
	np = length(x0)
	xs = zeros(np,niter+1)
	xs[:,1] = x0
end

fs = zeros(niter+1,1)
fs[1] = fcostold

out = Array{Any}(undef, niter+1)
out[1] = fun(x0, 0)

if mom == 0 && mu != 0
	alpha = 2. / (L+mu)
else
	alpha = 1. / L
end

ynew = []

# iterations
for iter=1:niter

#	@printf "iter=%d\n" iter (use "fun" for this)

	# gradient update
	fgrad = f_grad(xold)
	ynew = xold - alpha * fgrad # primary sequence
	fcostnew = f_cost(ynew)

	# restart + "gamma" decrease conditions
	if restart != 0

		# function/gradient restart
		if ((restart == 1 && fcostnew > fcostold)
				|| (restart == 2 && sum_double(-fgrad .* (ynew - yold)) < 0))
			told = 1
			sig = 1
			push!(rs, iter)

		# gradient "gamma" decrease
		elseif mom == 2 && sum_double(fgrad .* fgradold) < 0
			sig = bsig * sig
		end

		fcostold = fcostnew
		fgradold = fgrad
	end

	# momentum coefficient "beta"
	if mom == 1 && mu != 0 # known μ > 0
		beta = (1 - sqrt(q)) / (1 + sqrt(q))
	elseif mom == 2 && mu != 0
		beta = (2 + q - sqrt(q^2+8*q))^2 / 4. / (1-q)

	# for μ = 0 or for unknown μ
	elseif mom != 0
		if mom == 2 && iter == niter && restart == 0
			tnew = 0.5 * (1 + sqrt(1 + 8 * told^2))
		else
			tnew = 0.5 * (1 + sqrt(1 + 4 * told^2))
		end

		beta = (told - 1) / tnew
	end

	# momentum update
	if mom == 0
		xnew = ynew # GM
	elseif mom == 1
		xnew = ynew + beta * (ynew - yold) # FGM
	elseif mom == 2
		# momentum coefficient "gamma"
		if mu != 0
			gamma = (2 + q - sqrt(q^2+8*q)) / 2.
		else
			gamma = sig * told / tnew # OGM
		end

		xnew = ynew + beta * (ynew - yold) + gamma * (ynew - xold)
	end

	# save iterates
	if seq == 0
		fs[iter+1] = fcostnew
	elseif seq == 1
		xs[:,iter+1] = ynew
		fs[iter+1] = fcostnew
	elseif seq == 2
		xs[:,iter+1] = xnew
		fs[iter+1] = f_cost(xnew)
	else
		error("bad seq")
	end

	xold = xnew
	yold = ynew

	if mom != 0 && mu == 0
		told = tnew
	end

	if seq == 2
		out[iter+1] = fun(xnew, iter)
	else
		out[iter+1] = fun(ynew, iter)
	end
end # for iter

if seq == 0 # recommended way saves just last iterate
	xs = ynew
end

return (xs, fs, rs, out)
end # gm_restart()
