"""
sum with double precision
"""
function sum_double(x)
	return sum(Array{Float64}(x))
end

"""
	function pgm_restart(x0, F_cost, f_grad, f_L;
		f_mu=0, mom=0, restart=0, bsig=1, niter=10, g_prox=(z,c)->z)

Iterative proximal algorithms (PGM=ISTA, FPGM=FISTA, POGM) with restart

# Arguments
* `x0`: initial guess
* `F_cost`: function for computing the cost function value F(x)
* `f_grad`: function for computing the gradient of f(x)
* `f_L`: Lipschitz constant of the gradient of f(x)

# Options
* `f_mu`:	strong convexity parameter of f(x)
	if f_mu > 0, (alpha, beta_k, gamma_k) is chosen based on Table 1 in [KF18]
* `g_prox`:	function g_prox(z,c) for the proximal operator for g(x)
		"g_prox(z,c)" computes "argmin_x {1/2 ||z-x||^2 + c * g(x)}"
* `mom`:	momentum option
	0: PGM (ISTA) - beta_k = gamma_k = 0
	1: FPGM (FISTA) - gamma_k = 0
	2: POGM
* `restart`: restart option
		0: no restart
		1: function restart
		2: gradient restart
* `bsig`:	gradient "gamma" decrease option (value within [0 1]),
		see σ̄ (&sigma;&#772;) in [KF18]
* `niter`:	number of iterations

# Out
* `xs`:	iterates
		FPGM (FISTA) - primary iterates {y_k}
		POGM - secondary iterate {x_k}, see [KF18]
* `Fs`:	cost function values
* `rs`:	iterations where restarted

Optimization Problem: Nonsmooth Composite Convex Minimization
	min_x F(x) (:= f(x) + g(x))
		f(x): smooth convex function
		g(x): convex function, possibly nonsmooth and "proximal-friendly" [CP11]

Optimization Algorithms: Accelerated First-order Algorithms when "g(x) = 0" [KF18]
iterate as below for given coefficients "(alpha, beta_k, gamma_k)"
For k = 0,1,...
	y_{k+1} = x_k - alpha * f'(x_k) : gradient update
	x_{k+1} = y_{k+1} + beta_k * (y_{k+1} - y_k) + gamma_k * (y_{k+1} - x_k) : momentum update

The proximal version of the above for "g(x) ~= 0" can be found in the below references,
which uses the proximal operater "prox_g(z) = argmin_x {1/2||z-x||^2 + g(x)}"	

Proximal Gradient method (PGM or ISTA) - beta_k = gamma_k = 0. [BT09]
Fast Proximal Gradient Method (FPGM or FISTA) - gamma_k = 0. [BT09]
Proximal Optimized Gradient Method (POGM) - [THG15]
FPGM(FISTA) with Restart - [OC15]
POGM with Restart - [KF18]

* [CP11] P.L., Combettes, J.C. Pesquet,
"Proximal splitting methods in signal processing,"
Fixed-Point Algorithms for Inverse Problems in Science and Engineering,
Springer, Optimization and Its Applications, 2011.
* [KF18] D. Kim, J.A. Fessler,
"Adaptive restart of the optimized gradient method for convex optimization," 2018
Arxiv:1703.04641,
http://doi.org/10.1007/s10957-018-1287-4
* [BT09] A. Beck, M. Teboulle:
"A fast iterative shrinkage-thresholding algorithm for linear inverse problems,"
SIAM J. Imaging Sci., 2009.
* [THG15] A.B. Taylor, J.M. Hendrickx, F. Glineur,
"Exact worst-case performance of first-order algorithms
for composite convex optimization," Arxiv:1512.07516, 2015.

Copyright 2017-3-31, Donghwan Kim and Jeff Fessler, University of Michigan
2018-08-13 Julia 0.7.0
"""
function pgm_restart(x0, F_cost, f_grad, f_L;
	f_mu=0, mom=0, restart=0, bsig=1, niter=10, g_prox=(z,c)->z)

L = f_L
mu = f_mu
q = mu/L
rs = []

# initialize parameters
told = 1
sig = 1
zetaold = 1 # dummy

# initialize x
xold = x0
yold = x0
uold = x0
zold = x0
Fcostold = F_cost(x0)
Fgradold = zeros(size(x0)) # dummy

# save initial
np = length(x0)
xs = zeros(np,niter+1)
xs[:,1] = x0
Fs = zeros(niter+1)
Fs[1] = Fcostold

# iterations
for iter=1:niter

	# proximal gradient update
	if mom == 0 && mu != 0
		alpha = 2. / (L+mu)
	else
		alpha = 1. / L
	end

	fgrad = f_grad(xold)

	if mom < 2 # PGM, FPGM
		ynew = g_prox(xold - alpha * fgrad, alpha) # standard proximal gradient update
		Fgrad = -1. / alpha * (ynew - xold) # standard composite gradient mapping
		Fcostnew = F_cost(ynew)
		# restart condition
		if restart > 0
			# function/gradient restart
			if ((restart == 1 && Fcostnew > Fcostold)
			|| (restart == 2 && sum_double(-Fgrad .* (ynew - yold)) < 0))
				told = 1
				push!(rs, iter)
			end
			Fcostold = Fcostnew
		end

	else # POGM
		# gradient update for POGM [see KF18]
		unew = xold - alpha * fgrad
		# restart + "gamma" decrease conditions checked later for POGM,
		# unlike PGM, FPGM above
	end

	# momentum coefficient "beta"
	if mom == 1 && mu != 0 # known μ > 0
		beta = (1 - sqrt(q)) / (1 + sqrt(q))
	elseif mom == 2 && mu != 0
		beta = (2 + q - sqrt(q^2+8*q))^2 / 4. / (1-q)
	# for "mu" = 0 or for unknown "mu"
	elseif mom != 0
		if mom == 2 && iter == niter && restart == 0
			tnew = 0.5 * (1 + sqrt(1 + 8 * told^2))
		else
			tnew = 0.5 * (1 + sqrt(1 + 4 * told^2))
		end

		beta = (told - 1) / tnew
	end

	# momentum update
	if mom == 0
		xnew = ynew;
	elseif mom == 1
		xnew = ynew + beta * (ynew - yold);
	elseif mom == 2 # see [KF18]
		# momentum coefficient "gamma"
		if mu != 0
			gamma = (2 + q - sqrt(q^2+8*q)) / 2.
		else
			gamma = sig * told / tnew
		end

		znew = (unew + beta * (unew - uold) + gamma * (unew - xold)
				- beta * alpha / zetaold * (xold - zold))
		zetanew = alpha * (1 + beta + gamma)
		xnew = g_prox(znew, zetanew) # non-standard proximal gradient update for POGM
		Fgrad = fgrad - 1/zetanew * (xnew - znew) # non-standard composite gradient mapping for POGM
		ynew = xold - alpha * Fgrad
		Fcostnew = F_cost(xnew)

		# restart + "gamma" decrease conditions for POGM
		if restart > 0
			# function/gradient restart
			if ((restart == 1 && Fcostnew > Fcostold)
			|| (restart == 2 && sum_double(-Fgrad .* (ynew - yold)) < 0))
				tnew = 1
				sig = 1
				push!(rs, iter)

			# gradient "gamma" decrease
			elseif sum_double(Fgrad .* Fgradold) < 0
				sig = bsig * sig
			end

			Fcostold = Fcostnew
			Fgradold = Fgrad
		end

		uold = unew
		zold = znew
		zetaold = zetanew
	end

	# save iterates
	if mom < 2
		xs[:,iter+1] = ynew
		Fs[iter+1] = Fcostnew
	else
		xs[:,iter+1] = xnew
		Fs[iter+1] = Fcostnew # todo check
	end

	xold = xnew
	yold = ynew

	if mom != 0 && mu == 0
		told = tnew
	end
end # for iter

return (xs, Fs, rs)
end # pgm_restart()
