%| demo1_str_cvx_case1.m
%| demo 1
%|	demonstrates a strongly convex quadratic example in Section 6.1.1 of [KF17a]
%|
%| [KF17a] D. Kim, J.A. Fessler,
%| "Adaptive restart of the optimized gradient method for convex optimization,"
%| Arxiv:1703.04641, (2017)
%|
%| Copyright 2017-3-31, Donghwan Kim and Jeff Fessler, University of Michigan

%clear all
%close all

% generate problem
rng(2) % results depend a bit on the seed; unknown seed used in original paper plot
n = 500;
M = randn(n,n);
A = M'*M;
eps = 10^-4;
A = (1-eps) * A/max(eig(A)) + eps*eye(n,n);
b = randn(n,1);

% anonymous function for computing first-order information
f_cost = @(x) x'*(1/2*A*x - b);
f_grad = @(x) A*x - b;
% function parameters
f_L = max(eig(A));
f_mu = min(eig(A));

% number of iterations
niter = 1500;

% find optimum
xtmp = A \ b;
mom = 1; restart = 2; bsig = 1; seq = 1;
for i=1:10
	tmp = gm_restart(xtmp, f_cost, f_grad, f_L, f_mu, mom, restart, bsig, niter, seq);
	xtmp = tmp(:,end);
end
xopt = xtmp;
fopt = f_cost(xopt);

% generate initial
x0 = randn(n,1);

% GM
mom = 0; restart = 0; bsig = 1; seq = 1;
[xs, fs] = gm_restart(x0, f_cost, f_grad, f_L, 0, mom, restart, bsig, niter, seq);
[xsq, fsq] = gm_restart(x0, f_cost, f_grad, f_L, f_mu, mom, restart, bsig, niter, seq);

% FGM w/o restart
mom = 1;
[xsf, fsf] = gm_restart(x0, f_cost, f_grad, f_L, 0, mom, restart, bsig, niter, seq);
[xsfq, fsfq] = gm_restart(x0, f_cost, f_grad, f_L, f_mu, mom, restart, bsig, niter, seq);

% OGM w/o restart
mom = 2;
[xso, fso] = gm_restart(x0, f_cost, f_grad, f_L, 0, mom, restart, bsig, niter, seq);
[xsoq, fsoq] = gm_restart(x0, f_cost, f_grad, f_L, f_mu, mom, restart, bsig, niter, seq);

% FGM/OGM w/ function restart
mom = 1; restart = 1;
[xsf_fr, fsf_fr, rsf_fr] = gm_restart(x0, f_cost, f_grad, f_L, 0, mom, restart, bsig, niter, seq);
mom = 2;
[xso_fr, fso_fr, rso_fr] = gm_restart(x0, f_cost, f_grad, f_L, 0, mom, restart, bsig, niter, seq);

% FGM/OGM w/ gradient restart
mom = 1; restart = 2;
[xsf_gr, fsf_gr, rsf_gr] = gm_restart(x0, f_cost, f_grad, f_L, 0, mom, restart, bsig, niter, seq);
mom = 2;
[xso_gr, fso_gr, rso_gr] = gm_restart(x0, f_cost, f_grad, f_L, 0, mom, restart, bsig, niter, seq);

% display
%figure(1)
iter = 0:niter;
f = semilogy( ...
        iter, fs - fopt, 'k-', ...
        iter, fsq - fopt, 'k-.', ...
        iter, fsf - fopt, 'b-', ...
        iter, fso - fopt, 'r-', ... 
        iter, fsfq - fopt, 'b-.', ...
        iter, fsoq - fopt, 'r-.', ...
	iter, fsf_fr - fopt, 'c:', ...
        iter, fsf_gr - fopt, 'c--', ... 
        iter, fso_fr - fopt, 'm:', ...
        iter, fso_gr - fopt, 'm--');
l = {'GM', 'GM-$q$', 'FGM', 'OGM', ...
	'FGM-$q$', ...
	'OGM-$q$', ...
	'FGM-FR', 'FGM-GR', ...
	'OGM-FR', 'OGM-GR'};

xlim([0 niter])
ylim([10^(-6) f_cost(x0)-fopt])

ms = 11; gf = 16; ff= 14;
set(f, 'MarkerSize', ms, {'LineWidth'}, {1 1 3 5 3 5 3 3 5 5}');
set(gca, 'fontsize', gf);
legend(f, l, 'Location', 'SouthWest', 'fontsize', ff, 'Interpreter', 'latex');
xlabel('Iteration ($k$)', 'Interpreter', 'latex');
ylabel('$f(y_k) - f(x_*)$', 'Interpreter', 'latex');

% save fig
% print('-depsc', '../fig/fig,demo1_str_cvx_case1.eps')
