%| demo3_logsumexp.m
%| demo 3
%| demonstrates two "Log-Sum-Exp" examples in Section 6.2.1 of [KF17a]
%|
%| [KF17a] D. Kim, J.A. Fessler,
%| "Adaptive restart of the optimized gradient method for convex optimization,"
%| Arxiv:1703.04641, (2017)
%|
%| Copyright 2017-3-31, Donghwan Kim and Jeff Fessler, University of Michigan

% generate problem
n = 20;
m = 100;
rng(0) % results depend a bit on the seed; unknown what seed was used originally
A = randn(m,n);
b = randn(m,1);

% problem parameter
rholist = [1 10];

% number of iterations
niter = 1000;

% two examples
for opt = 1:2
	rho = rholist(opt);

	% anonymous functions for computing first-order information
	f_cost = @(x) rho * log(sum(exp((A*x-b)/rho), 'double'));
	f_grad = @(x) 1/sum(exp((A*x-b)/rho), 'double') * (A' * exp((A*x-b)/rho));

	% function parameters
	f_L = max(eig(A'*A))/rho;
	f_mu = 0; % zero is a lower bound

	% find optimum
	xtmp = randn(n,1);
	mom = 1; restart = 2; bsig = 1; seq = 1;
	for iter=1:20
		tmp = gm_restart(xtmp, f_cost, f_grad, f_L, ...
			f_mu, mom, restart, bsig, niter, seq);
		xtmp = tmp(:,end);
	end
	xopt = xtmp;
	fopt = f_cost(xopt);

	% generate initial
	x0 = randn(n,1);

	% GM
	mom = 0; restart = 0; bsig = 1; seq = 1;
	[xs, fs] = gm_restart(x0, f_cost, f_grad, f_L, ...
			f_mu, mom, restart, bsig, niter, seq);

	% FGM w/o restart
	mom = 1;
	[xsf, fsf] = gm_restart(x0, f_cost, f_grad, f_L, ...
			f_mu, mom, restart, bsig, niter, seq);

	% OGM w/o restart
	mom = 2;
	[xso, fso] = gm_restart(x0, f_cost, f_grad, f_L, ...
			f_mu, mom, restart, bsig, niter, seq);

	% FGM/OGM w/ function restart
	mom = 1; restart = 1;
	[xsf_fr, fsf_fr, rsf_fr] = gm_restart(x0, f_cost, f_grad, f_L, ...
			f_mu, mom, restart, bsig, niter, seq);
	mom = 2;
	[xso_fr, fso_fr, rso_fr] = gm_restart(x0, f_cost, f_grad, f_L, ...
			f_mu, mom, restart, bsig, niter, seq);

	% FGM/OGM w/ gradient restart
	mom = 1; restart = 2;
	[xsf_gr, fsf_gr, rsf_gr] = gm_restart(x0, f_cost, f_grad, f_L, ...
			f_mu, mom, restart, bsig, niter, seq);
	mom = 2;
	[xso_gr, fso_gr, rso_gr] = gm_restart(x0, f_cost, f_grad, f_L, ...
			f_mu, mom, restart, bsig, niter, seq);

	% display
	figure(opt);
	iter = 0:niter;
	f = semilogy( ...
		iter, (fs - fopt)/fopt, 'k-', ...
		iter, (fsf - fopt)/fopt, 'b-', ...
		iter, (fso - fopt)/fopt, 'r-', ...
		iter, (fsf_fr - fopt)/fopt, 'c:', ...
		iter, (fsf_gr - fopt)/fopt, 'c--', ...
		iter, (fso_fr - fopt)/fopt, 'm:', ...
		iter, (fso_gr - fopt)/fopt, 'm--');
	l = {'GM', 'FGM', 'OGM', 'FGM-FR', 'FGM-GR', 'OGM-FR', 'OGM-GR'};

	xlim([0 niter])
	ylim([10^-15 (f_cost(x0) - fopt)/fopt])

	ms = 11; gf = 16; ff= 14;
	set(f, 'MarkerSize', ms, {'LineWidth'}, {1 3 5 3 3 5 5}');
	set(gca, 'fontsize', gf);
	if opt == 1
		legend(f, l, 'Location', 'SouthWest', 'fontsize', ff);
	end
	xlabel('Iteration ($k$)', 'Interpreter', 'latex');
	ylabel('$(f(y_k) - f(x_*))/f(x_*)$', 'Interpreter', 'latex');
	title(sprintf('$\\eta=%d$', rho), 'Interpreter', 'latex');

	% save fig
%	current_path = pwd;
	fig_path = sprintf('../fig/fig,demo3_logsumexp%d.eps', opt);
%	print('-depsc', fig_path)
end
