%| demo4_sparselr
%| demo 4
%| Demonstrates two "Sparse Linear Regression" (LASSO) examples
%| in Fig. 5 of Section 6.2.2 of [KF17a]
%|
%| [KF17a] D. Kim, J.A. Fessler,
%| "Adaptive restart of the optimized gradient method for convex optimization,"
%| Arxiv:1703.04641, (2017)
%|
%| Copyright 2017-3-31, Donghwan Kim and Jeff Fessler, University of Michigan

%clear all;
%close all;
rng(1) % unknown what seed was used originally for paper

color_peach = [251 111 66] ./ 255; % for plot

% problem parameters
n = 2000;
mlist = [500 100];
slist = [100 20];
rholist = [1 10];

% number of iterations
niter = 4000;

for opt=1:2
	m = mlist(opt);
	s = slist(opt);
	rho = rholist(opt);

	% generate problem
	A = randn(m,n);
	xtmp = randn(n,1);
	[tmp, indx] = sort(abs(xtmp), 'descend');
	xtmp(indx(s+1:end)) = 0;
	xtrue = xtmp;
%	pr sum(xtrue == 0) % # of zeros
	b = A*xtrue + sqrt(0.1)*randn(m,1);

	% anonymous functions for computing first-order information
	F_cost = @(x) 1/2*norm(A*x-b,2)^2 + rho*norm(x,1);
	f_grad = @(x) A'*(A*x-b);
	g_prox = @(x, c) sign(x) .* max(abs(x)-c*rho, 0);
	% function parameters
	f_L = max(eig(A'*A));
	f_mu = 0; % zero is a lower bound

	% find optimum
	xtmp = xtrue;
%	pr F_cost(xtmp) % initial cost
	mom = 1; restart = 2; bsig = 1;
	for outer_iter = 1:5
		tmp = pgm_restart(xtmp, F_cost, f_grad, f_L, f_mu, ...
			g_prox, mom, restart, bsig, niter);
		xtmp = tmp(:,end);
%		pr F_cost(xtmp)
	end
	xopt = xtmp;
	Fopt = F_cost(xopt);
%	pr Fopt

	% generate initial
	x0 = randn(n,1);

	% ISTA
	mom = 0; restart = 0; bsig = 1;
	[xs, Fs] = pgm_restart(x0, F_cost, f_grad, f_L, f_mu, ...
		g_prox, mom, restart, bsig, niter);

	% FISTA w/o restart
	mom = 1;
	[xsf, Fsf] = pgm_restart(x0, F_cost, f_grad, f_L, f_mu, ...
		g_prox, mom, restart, bsig, niter);

	% POGM w/o restart
	mom = 2;
	[xso, Fso] = pgm_restart(x0, F_cost, f_grad, f_L, f_mu, ...
		g_prox, mom, restart, bsig, niter);

	% FISTA/POGM w/ function restart
	mom = 1; restart = 1;
	[xsf_fr, Fsf_fr, rsf_fr] = pgm_restart(x0, F_cost, f_grad, f_L, f_mu, ...
		g_prox, mom, restart, bsig, niter);
	mom = 2;
	[xso_fr, Fso_fr, rso_fr] = pgm_restart(x0, F_cost, f_grad, f_L, f_mu, ...
		g_prox, mom, restart, bsig, niter);

	% FISTA/POGM w/ gradient restart
	mom = 1; restart = 2;
	[xsf_gr, Fsf_gr, rsf_gr] = pgm_restart(x0, F_cost, f_grad, f_L, f_mu, ...
		g_prox, mom, restart, bsig, niter);
	mom = 2;
	[xso_gr, Fso_gr, rso_gr] = pgm_restart(x0, F_cost, f_grad, f_L, f_mu, ...
		g_prox, mom, restart, bsig, niter);

	% POGM w/ gradient restart + gradient gamma decreasing (bsig = 0.8)
	bsig = 0.8;
	[xso_grg, Fso_grg, rso_grg] = pgm_restart(x0, F_cost, f_grad, f_L, f_mu, ...
		g_prox, mom, restart, bsig, niter);

	% display
	figure(opt)
	iter = 0:niter;
	f = semilogy( ...
		iter, (Fs - Fopt)/Fopt, 'k-', ...
		iter, (Fsf - Fopt)/Fopt, 'b-', ...
		iter, (Fso - Fopt)/Fopt, 'r-', ...
		iter, (Fsf_fr - Fopt)/Fopt, 'c:', ...
		iter, (Fsf_gr - Fopt)/Fopt, 'c--', ...
		iter, (Fso_fr - Fopt)/Fopt, 'm:', ...
		iter, (Fso_gr - Fopt)/Fopt, 'm--', ...
		iter, (Fso_grg - Fopt)/Fopt, 'g-.');
	l = {'ISTA', 'FISTA', 'POGM', 'FISTA-FR', 'FISTA-GR', ...
		'POGM-FR-GD$\gamma$($\bar{\sigma} = 1.0$)', ...
		'POGM-GR-GD$\gamma$($\bar{\sigma} = 1.0$)', ...
		'POGM-GR-GD$\gamma$($\bar{\sigma} = 0.8$)'};
	set(f(8), 'color', color_peach);

	xlim([0 niter])
	ylim([10^-13 (F_cost(x0) - Fopt)/Fopt])

	ms = 11; gf = 16; ff= 14;
	set(f, 'MarkerSize', ms, {'LineWidth'}, {1 3 5 3 3 5 5 3}')
	set(gca, 'fontsize', gf)
	if opt == 1
		legend(f, l, 'Location', 'NorthEast', 'fontsize', ff, 'Interpreter', 'latex')
	end
	xlabel('Iteration ($k$)', 'Interpreter', 'latex')
	ylabel('$(F(y_k) - F(x_*))/F(x_*)$ or $(F(x_k) - F(x_*))/F(x_*)$', 'Interpreter', 'latex')
	title(sprintf('$m=%d,\\;d=%d,\\;s=%d,\\;\\tau=%d$', m, n, s, rho), 'Interpreter', 'latex')

	% save fig
	fig_path = sprintf('../fig/fig,demo4_sparselr%d.eps', opt);
%	print('-depsc', fig_path)
end
