%| demo5_quad.m
%| demo 5
%|	demonstrates two "Constrained Quadratic Programming" examples
%|	in Section 6.2.3 of [KF17a]
%|
%| These figures are in the arxiv version only, not in the published JOTA version.
%|
%| [KF17a] D. Kim, J.A. Fessler
%| "Adaptive restart of the optimized gradient method for convex optimization,"
%| Arxiv:1703.04641, (2017)
%|
%| Copyright 2017-3-31, Donghwan Kim and Jeff Fessler, University of Michigan

rng(0) % unknown what seed was used in original figures

MaxNum = realmax('double'); % simple alternative to infinity for cost function

nlist = [500 1000]; % problem parameter
niter = 2000; % number of iterations
pr = @(l) disp(l); % print status

for opt = 1:2
	n = nlist(opt);

	% generate problem
	A = randn(n,n);
	Q = A'*A;
	eps = 10^-7;
	Q = (1-eps) * Q/max(eig(Q)) + eps*eye(n,n);
	b = 0.005 * randn(n,1); % todo: where does constant originate?
	lb = -ones(n,1);
	ub = ones(n,1);

	% anonymous functions for computing first-order information
	F_cost = @(x) x'*(1/2*Q*x - b) + MaxNum*(sum(x<lb | x>ub));
	f_grad = @(x) Q*x-b;
	g_prox = @(x, c) min(max(x, lb), ub);
	% function parameters
	f_L = max(eig(Q));
	f_mu = 0; % zero is a lower bound

	% find optimum
	xtmp = g_prox(Q\b, 1);
	mom = 1; restart = 2; bsig = 1;
	for i=1:7
		tmp = pgm_restart(xtmp, F_cost, f_grad, f_L, f_mu, ...
			g_prox, mom, restart, bsig, niter);
		xtmp = tmp(:,end);
	end
	xopt = xtmp;
	Fopt = F_cost(xopt);

	% generate initial
	x0 = g_prox(randn(n,1), 1);

	% ISTA
	pr('ista')
	mom = 0; restart = 0; bsig = 1;
	[xs, Fs] = pgm_restart(x0, F_cost, f_grad, f_L, f_mu, ...
		g_prox, mom, restart, bsig, niter);

	% FISTA w/o restart
	pr('fista')
	mom = 1;
	[xsf, Fsf] = pgm_restart(x0, F_cost, f_grad, f_L, f_mu, ...
		g_prox, mom, restart, bsig, niter);

	% POGM w/o restart
	pr('pogm')
	mom = 2;
	[xso, Fso] = pgm_restart(x0, F_cost, f_grad, f_L, f_mu, ...
		g_prox, mom, restart, bsig, niter);

	% FISTA/POGM w/ function restart
	pr('f restart')
	mom = 1; restart = 1;
	[xsf_fr, Fsf_fr, rsf_fr] = pgm_restart(x0, F_cost, f_grad, ...
		f_L, f_mu, g_prox, mom, restart, bsig, niter);
	mom = 2;
	[xso_fr, Fso_fr, rso_fr] = pgm_restart(x0, F_cost, f_grad, ...
		f_L, f_mu, g_prox, mom, restart, bsig, niter);

	% FISTA/POGM w/ gradient restart
	pr('g restart')
	mom = 1; restart = 2;
	[xsf_gr, Fsf_gr, rsf_gr] = pgm_restart(x0, F_cost, f_grad, ...
		f_L, f_mu, g_prox, mom, restart, bsig, niter);
	mom = 2;
	[xso_gr, Fso_gr, rso_gr] = pgm_restart(x0, F_cost, f_grad, ...
		f_L, f_mu, g_prox, mom, restart, bsig, niter);

	% POGM w/ gradient restart + gradient gamma decreasing (bsig = 0.8)
	pr('pogm')
	bsig = 0.8;
	[xso_grg, Fso_grg, rso_grg] = pgm_restart(x0, F_cost, f_grad, ...
		f_L, f_mu, g_prox, mom, restart, bsig, niter);

	% display
	figure(opt);
	iter = 0:niter;
	f = semilogy( ...
		iter, (Fs - Fopt), 'k-', ...
		iter, (Fsf - Fopt), 'b-', ...
		iter, (Fso - Fopt), 'r-', ...
		iter, (Fsf_fr - Fopt), 'c:', ...
		iter, (Fsf_gr - Fopt), 'c--', ...
		iter, (Fso_fr - Fopt), 'm:', ...
		iter, (Fso_gr - Fopt), 'm--', ...
		iter, (Fso_grg - Fopt), 'g-.');
	l = {'Proj. GM', 'Proj. FGM', 'Proj. OGM', ...
		'Proj. FGM-FR', 'Proj. FGM-GR', ...
		'Proj. OGM-FR-GD$\gamma$($\bar{\sigma} = 1.0)$', ...
		'Proj. OGM-GR-GD$\gamma$($\bar{\sigma} = 1.0)$', ...
		'Proj. OGM-GR-GD$\gamma$($\bar{\sigma} = 0.8)$'};
	color_peach = [251 111 66] ./ 255; % for plot
	set(f(8), 'color', color_peach);

	xlim([0 niter])
	ylim([10^-12 (F_cost(x0) - Fopt)])

	ms = 11; gf = 16; ff= 14;
	set(f, 'MarkerSize', ms, {'LineWidth'}, {1 3 5 3 3 5 5 3}');
	set(gca, 'fontsize', gf);
	if opt == 1
		legend(f, l, 'Location', 'NorthEast', 'fontsize', ff, 'Interpreter', 'latex');
	end
	xlabel('Iteration ($k$)', 'Interpreter', 'latex');
	ylabel('$F(y_k) - F(x_*)$ or $F(x_k) - F(x_*)$', 'Interpreter', 'latex');
	title(sprintf('$d=%d\\;$ ($x_*$ has %d active constraints)', ...
		n, sum(xopt==lb | xopt==ub)), 'Interpreter', 'latex');

	% save fig
	fig_path = sprintf('../fig,demo5_quad%d_n%d.eps', opt, n);
%	print('-depsc', fig_path)
end
