 function [xs, fs, rs] = gm_restart(x0, f_cost, f_grad, f_L, f_mu, mom, restart, bsig, niter, seq)
%function [xs, fs, rs] = gm_restart(x0, f_cost, f_grad, f_L, f_mu, mom, restart, bsig, niter, seq)
%|
%| Optimization Problem: Smooth Convex Minimization
%|	min_x f(x)
%|
%| Optimization Algorithms: Accelerated First-order Algorithms [KF17a]
%|	iterate as below for given coefficients "(alpha, beta_k, gamma_k)"
%|	For k = 0,1,...
%|		y_{k+1} = x_k - alpha * f'(x_k) : gradient update
%|		x_{k+1} = y_{k+1} + beta_k * (y_{k+1} - y_k) + gamma_k * (y_{k+1} - x_k)
%|									: momentum update
%|	Gradient method (GM) : beta_k = gamma_k = 0.
%|	Fast Gradient Method (FGM) : gamma_k = 0. [N83, BT09]
%|	Optimized Gradient Method (OGM) : [KF16] [KF17b]
%|	FGM with Restart : [OC15]
%|	OGM with Restart : [KF17a]
%|
%| [KF17a] D. Kim, J.A. Fessler, "Adaptive restart of the optimized gradient method for convex
%|	optimization," Arxiv:1703.04641, (2017)
%| [N83] Y. Nesterov, "A method for unconstrained convex minimization problem with the rate of
%|	convergence O(1/k^2)," Dokl. Akad. Nauk. USSR, (1983).
%| [BT09] A. Beck, M. Teboulle: "A fast iterative shrinkage-thresholding algorithm for linear
%|	inverse problems," SIAM J. Imaging Sci., (2009).
%| [KF16] D. Kim, J.A. Fessler, "Optimized first-order methods for smooth convex minimization,"
%|	Math. Program., (2016).
%| [KF17b] D. Kim, J.A. Fessler, "On the convergence analysis of the optimized gradient method,"
%|	J. Optim. Theory Appl., (2017).
%| [OC15] B. O’Donoghue, E. Candes, "Adaptive restart for accelerated gradient schemes,"
%|	Found. Comp. Math., (2015).
%|
%| input
%|	x0	initial
%|	f_cost	inline function for computing the cost function value f(x)
%|	f_grad	inline function for computing the gradient of f(x)
%|	f_L	Lipschitz constant of the gradient of f(x)
%|	f_mu	strong convexity parameter of f(x)
%|		if f_mu > 0, (alpha, beta_k, gamma_k) is chosen based on Table 1 in [KF17a]
%|	mom	momentum option
%|		0: GM - beta_k = gamma_k = 0
%|		1: FGM - gamma_k = 0
%|		2: OGM
%|	restart restart option
%|		0: no restart
%|		1: function restart, see [OC15, KF17a]
%|		2: gradient restart, see [OC15, KF17a]
%|	bsig	gradient "gamma" decrease option (value within [0 1]),
%|		see "\bar{\sigma}" in [KF17a]
%|	niter	number of iterations
%|	seq	output iterates
%|		1: primary iterates {y_k}
%|		2: secondary iterates {x_k}
%|
%| output
%|	xs	iterates
%|	fs	cost function values
%|	rs	restarted points
%|
%| Copyright 2017-3-31, Donghwan Kim and Jeff Fessler, University of Michigan

L = f_L;
mu = f_mu;
q = mu/L;
rs = [];

% initialize parameters
told = 1;
sig = 1;

% initialize x
xold = x0;
yold = x0;
fcostold = f_cost(x0);
fgradold = zeros(size(x0)); % dummy

% save initial
xs = [x0];
fs = [fcostold];

% iterations
for n=1:niter
	% gradient update
	if mom == 0 && mu
		alpha = 2/(L+mu);
	else
		alpha = 1/L;
	end
	fgrad = f_grad(xold);
	ynew = xold - alpha * fgrad;
	fcostnew = f_cost(ynew);

	% restart + "gamma" decrease conditions
	if restart
		% function/gradient restart
		if (restart == 1 && fcostnew > fcostold) ...
			|| (restart == 2 && sum(-fgrad .* (ynew - yold), 'double') < 0)
			told = 1;
			sig = 1;
			rs(end+1) = n;
		% gradient "gamma" decrease
		elseif mom == 2 && sum(fgrad .* fgradold, 'double') < 0
			sig = bsig * sig;
		end
		fcostold = fcostnew;
		fgradold = fgrad;
	end

	% momentum coefficient "beta"
	% with known "mu" > 0
	if mom == 1 && mu
		beta = (1 - sqrt(q))/(1 + sqrt(q));
	elseif mom == 2 && mu
		beta = (2 + q - sqrt(q^2+8*q))^2 / 4 / (1-q);
	% for "mu" = 0 or for unknown "mu"
	elseif mom
		if mom == 2 && n == niter && restart == 0
			tnew = 1/2 * (1 + sqrt(1 + 8 * told^2));
		else
			tnew = 1/2 * (1 + sqrt(1 + 4 * told^2));
		end
		beta = (told - 1) / tnew;	
	end

	% momentum update
	if mom == 0
		xnew = ynew;
	elseif mom == 1
		xnew = ynew + beta * (ynew - yold);
	elseif mom == 2
		% momentum coefficient "gamma"
		if mu
			gamma = (2 + q - sqrt(q^2+8*q)) / 2;
		else
			gamma = sig * told / tnew;
		end
		xnew = ynew + beta * (ynew - yold) + gamma * (ynew - xold);
	end
	
	% save iterates
	if seq == 1
		xs = [xs, ynew];
		fs = [fs, fcostnew];
	else
		xs = [xs, xnew];
		fs = [fs, f_cost(xnew)];
	end
	xold = xnew;
	yold = ynew;
	if mom && mu == 0
		told = tnew;
	end
end
