 function [xs, Fs, rs] = pgm_restart(x0, F_cost, f_grad, f_L, f_mu, g_prox, mom, restart, bsig, niter)
%function [xs, Fs, rs] = pgm_restart(x0, F_cost, f_grad, f_L, f_mu, g_prox, mom, restart, bsig, niter)
%|
%| Optimization Problem: Nonsmooth Composite Convex Minimization
%|	min_x F(x) (:= f(x) + g(x))
%|		f(x): smooth convex function
%|		g(x): convex function, possibly nonsmooth and "proximal-friendly" [CP11]
%|
%| Optimization Algorithms: Accelerated First-order Algorithms when "g(x) = 0" [KF17a]
%|	iterate as below for given coefficients "(alpha, beta_k, gamma_k)"
%|	For k = 0,1,...
%|		y_{k+1} = x_k - alpha * f'(x_k) : gradient update
%|		x_{k+1} = y_{k+1} + beta_k * (y_{k+1} - y_k) + gamma_k * (y_{k+1} - x_k) : momentum update
%|	The proximal version of the above for "g(x) ~= 0" can be found in the below references,
%|	which uses the proximal operater "prox_g(z) = argmin_x {1/2||z-x||^2 + g(x)}"	
%|
%|	Proximal Gradient method (PGM or ISTA) - beta_k = gamma_k = 0. [BT09]
%|	Fast Proximal Gradient Method (FPGM or FISTA) - gamma_k = 0. [BT09]
%|	Proximal Optimized Gradient Method (POGM) - [THG15]
%|	FPGM(FISTA) with Restart - [OC15]
%|	POGM with Restart - [KF17a]
%|
%| [CP11] P.L., Combettes, J.C. Pesquet, "Proximal splitting methods in signal processing,"
%|	Fixed-Point Algorithms for Inverse Problems in Science and Engineering, Springer,
%|	Optimization and Its Applications, (2011).
%| [KF17a] D. Kim, J.A. Fessler, "Adaptive restart of the optimized gradient method for convex
%|	optimization," Arxiv:1703.04641, (2017)
%| [BT09] A. Beck, M. Teboulle: "A fast iterative shrinkage-thresholding algorithm for linear
%|	inverse problems," SIAM J. Imaging Sci., (2009).
%| [THG15] A.B. Taylor, J.M. Hendrickx, F. Glineur, "Exact worst-case performance of first-order
%|	algorithms for composite convex optimization," Arxiv:1512.07516, (2015).
%|
%| input
%|	x0	initial
%|	F_cost	inline function for computing the function value F(x)
%|	f_grad	inline function for computing the gradient of f(x)
%|	f_L	Lipschitz constant of the gradient of f(x)
%|	f_mu	strong convexity parameter of f(x)
%|		if f_mu > 0, (alpha, beta_k, gamma_k) is chosen based on Table 1 in [KF17a]
%|	g_prox	inline function g_prox(x,c) for the proximal operator for g(x)
%|		"g_prox(x,c)" computes "argmin_x {1/2||z-x||^2 + c * g(x)}"
%|	mom	momentum option
%|		0: PGM (ISTA) - beta_k = gamma_k = 0
%|		1: FPGM (FISTA) - gamma_k = 0
%|		2: POGM
%|	restart restart option
%|		0: no restart
%|		1: function restart
%|		2: gradient restart
%|	bsig	gradient "gamma" decrease option (value within [0 1]),
%|		see "\bar{\sigma}" in [KF17a]
%|	niter	number of iterations
%|
%| out
%|	xs	iterates
%|		FPGM (FISTA) - primary iterates {y_k}
%|		POGM - secondary iterate {x_k}, see [KF17a]
%|	Fs	cost function values
%|	rs	restarted points
%|
%| Copyright 2017-3-31, Donghwan Kim and Jeff Fessler, University of Michigan

L = f_L;
mu = f_mu;
q = mu/L;
rs = [];

% initialize parameters
told = 1;
sig = 1;
zetaold = 1; % dummy

% initialize x
xold = x0;
yold = x0;
uold = x0;
zold = x0;
Fcostold = F_cost(x0);
Fgradold = zeros(size(x0)); % dummy

% save initial
xs = [x0];
Fs = [Fcostold];

% iterations
for n=1:niter
	% proximal gradient update
	if mom == 0 && mu
		alpha = 2/(L+mu);
	else
		alpha = 1/L;
	end
	fgrad = f_grad(xold);
	if mom < 2 % PGM, FPGM
		ynew = g_prox(xold - alpha * fgrad, alpha); % standard proximal gradient update
		Fgrad = - 1/alpha * (ynew - xold); % standard composite gradient mapping
		Fcostnew = F_cost(ynew);
		% restart condition
		if restart
			% function/gradient restart
			if (restart == 1 && Fcostnew > Fcostold) ...
			|| (restart == 2 && sum(-Fgrad .* (ynew - yold), 'double') < 0)
				told = 1;
				rs(end+1) = n;
			end
			Fcostold = Fcostnew;
		end

	else % POGM
		% gradient update for POGM [see KF17a]
		unew = xold - alpha * fgrad;
		% restart + "gamma" decrease conditions checked later for POGM,
		% unlike PGM, FPGM above
	end

	% momentum coefficient "beta"
	% with known "mu" > 0
	if mom == 1 && mu
		beta = (1 - sqrt(q))/(1 + sqrt(q));
	elseif mom == 2 && mu
		beta = (2 + q - sqrt(q^2+8*q))^2 / 4 / (1-q);
	% for "mu" = 0 or for unknown "mu"
	elseif mom
		if mom == 2 && n == niter && restart == 0
			tnew = 1/2 * (1 + sqrt(1 + 8 * told^2));
		else
			tnew = 1/2 * (1 + sqrt(1 + 4 * told^2));
		end
		beta = (told - 1) / tnew;	
	end

	% momentum update
	if mom == 0
		xnew = ynew;
	elseif mom == 1
		xnew = ynew + beta * (ynew - yold);
	elseif mom == 2 % see [KF17a]
		% momentum coefficient "gamma"
		if mu
			gamma = (2 + q - sqrt(q^2+8*q)) / 2;
		else
			gamma = sig * told / tnew;
		end
		znew = unew + beta * (unew - uold) + gamma * (unew - xold) ...
				- beta * alpha / zetaold * (xold - zold);
		zetanew = alpha * (1 + beta + gamma);
		xnew = g_prox(znew, zetanew); % non-standard proximal gradient update for POGM
		Fgrad = fgrad - 1/zetanew * (xnew - znew); % non-standard composite gradient mapping for POGM
		ynew = xold - alpha*Fgrad;
		Fcostnew = F_cost(xnew);
		% restart + "gamma" decrease conditions for POGM
		if restart
			% function/gradient restart
			if (restart == 1 && Fcostnew > Fcostold) ...
			|| (restart == 2 && sum(-Fgrad .* (ynew - yold), 'double') < 0)
				tnew = 1;
				sig = 1;
				rs(end+1) = n;
			% gradient "gamma" decrease
			elseif sum(Fgrad .* Fgradold, 'double') < 0
				sig = bsig * sig;
			end
			Fcostold = Fcostnew;
			Fgradold = Fgrad;
		end
		uold = unew;
		zold = znew;
		zetaold = zetanew;
	end
	
	% save iterates
	if mom < 2
		xs = [xs, ynew];
		Fs = [Fs, Fcostnew];
	else
		xs = [xs, xnew];
		Fs = [Fs, Fcostnew];
	end
	xold = xnew;
	yold = ynew;
	if mom && mu == 0
		told = tnew;
	end
end
